#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    _Back.py

    The inverse Radon transform with non-iterative techniques.
"""

from __future__ import division

import numpy as np
import os
import scipy.interpolate as intp
import warnings

__all__= ["backproject", "fourier_map", "sum"]


def backproject(sinogram, angles, filtering="ramp",
                jmc=None, jmm=None):
    u""" 2D backprojection with the Fourier slice theorem
    
    Computes the inverse of the radon transform using filtered
    backprojection.


    Parameters
    ----------
    sinogram : ndarray, shape (A,N)
        Two-dimensional sinogram of line recordings.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI.
    filtering : {'ramp', 'shepp-logan', 'cosine', 'hamming', \
                 'hann'}, optional
        Specifies the Fourier filter. Either of
        
        ``ramp``
          mathematically correct reconstruction
          
        ``shepp-logan``
        
        ``cosine``
        
        ``hamming``
        
        ``hann``
        
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.


    Returns
    -------
    out : ndarray
        The reconstructed image.


    Examples
    --------
    >>> import numpy as np
    >>> from radontea import backproject
    >>> N = 5
    >>> A = 7
    >>> x = np.linspace(-N/2, N/2, N)
    >>> projection = np.exp(-x**2)
    >>> sinogram = np.tile(projection, A).reshape(A,N)
    >>> angles = np.linspace(0, np.pi, A, endpoint=False)
    >>> backproject(sinogram, angles)
    array([[ 0.03813283, -0.01972347, -0.02221885,  0.03822303,  0.01903376],
           [ 0.04033526, -0.01591647,  0.02262173,  0.04203002,  0.02123619],
           [ 0.02658797,  0.11375576,  0.41525594,  0.17170226,  0.0074889 ],
           [ 0.04008672,  0.11139209,  0.27650193,  0.16933859,  0.02098764],
           [ 0.02140445,  0.0334597 ,  0.07691547,  0.0914062 ,  0.00230537]])


    See Also
    --------
    fourier_map : implementation by Fourier interpolation
    sum : implementation by summation in real space
    """
    ln = len(sinogram[0])
    la = len(angles)
    # jobmanager init
    if jmm is not None:
        jmm.value = la+1
    # transpose so we can call resize correctly
    sino = sinogram.transpose().copy()
    # Apply a Fourier filter before projecting the sinogram slices.
    # Resize image to next power of two for fourier analysis
    # Speeds up fourier and lessens artifacts
    order = max(64., 2**np.ceil(np.log(2 * ln) / np.log(2)))
    sino.resize((order,la))
    sino = sino.transpose()
    # These artifacts are for example bad contrast. To check, set:
    #order = ln
    #sino = sinogram
    kx = 2 * np.pi * np.abs(np.fft.fftfreq(int(order)))
    # Ask for the filter. Do not include zero (first element).
    if filtering == "ramp":
        pass
    elif filtering == "shepp-logan":
        kx[1:] = kx[1:] * np.sin(kx[1:]) / (kx[1:])
    elif filtering == "cosine":
        kx[1:] = kx[1:] * np.cos(kx[1:])
    elif filtering == "hamming":
        kx[1:] = kx[1:] * (0.54 + 0.46 * np.cos(kx[1:]))
    elif filtering == "hann":
        kx[1:] = kx[1:] * (1 + np.cos(kx[1:])) / 2
    elif filtering == None:
        kx[1:] = 2*np.pi
    else:
        raise ValueError("Unknown filter: %s" % filter)
    # jobmanager
    if jmc is not None:
        jmc.value += 1
    # Resize f so we can multiply it with the sinogram.
    kx = kx.reshape(1,-1)
    projection = np.fft.fft(sino, axis=-1) * kx
    sino_filtered = np.real(np.fft.ifft(projection, axis=-1))
    # Resize filtered sinogram back to original size
    sino = sino_filtered[:, :ln]
    ## Perform backprojection
    x = np.linspace(-ln/2, ln/2, ln, endpoint=False) +.5
    outarr = np.zeros((ln,ln),dtype=np.float64)
    # Meshgrid for output array
    xv, yv = np.meshgrid(x,x)
        
    for i in np.arange(len(angles)):
        #xproj = np.linspace(-n/2.0, n/2.0, N, endpoint=False)
        projinterp = intp.interp1d(x, sino[i], fill_value=0.0,
                                   copy=False, bounds_error=False)
        # Call proj_interp with cos and sin of corresponding coordinate
        # and add it to the outarr.
        phi = angles[i]
        # Smear projection onto 2d volume
        xp = xv*np.cos(phi) + yv*np.sin(phi)
        outarr += projinterp(xp)
        ## Here a much slower version with the same result:
        #for j in x:
        #    for k in x:
        #        # shift points to match origin of coordinate system
        #        xv = j - ln/2
        #        yv = k - ln/2
        #        # perform a simple coordinate transform
        #        # 
        #        xp = xv*np.cos(phi) + yv*np.sin(phi)
        #        # yp = -xv*np.sin(phi) + yv*np.cos(phi)
        #        projval = projinterp(xp)
        #        outarr[j][k] += projval
        if jmc is not None:
            jmc.value += 1
    # Normalize output (we assume that the projections are equidistant)
    # We measure angles in degrees
    dphi = np.pi/len(angles)
    # The factor of 2π comes from the choice of the unitary angular
    # frequency Fourier transform.
    outarr *= dphi / (2*np.pi)

    return outarr


def fourier_map(sinogram, angles, intp_method="cubic",
                   jmc=None, jmm=None):
    u""" 2D Fourier mapping with the Fourier slice theorem
    
    Computes the inverse of the radon transform using Fourier
    interpolation.
    Warning: This is the naive reconstruction that assumes that
    the image is rotated through the upper left pixel nearest to
    the actual center of the image. We do not have this problem for
    odd images, only for even images.

    
    Parameters
    ----------
    sinogram : (A,N) ndarray
        Two-dimensional sinogram of line recordings.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI.
    intp_method : {'cubic', 'nearest', 'linear'}, optional
        Method of interpolation. For more information see
        `scipy.interpolate.griddata`. One of
        
        ``nearest``
          instead of interpolating, use the points closest to
          the input data.

        ``linear``
          bilinear interpolation between data points.
          
        ``cubic``
          interpolate using a two-dimensional poolynimial surface.
          
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.


    Returns
    -------
    out : ndarray
        The reconstructed image.


    See Also
    --------
    backproject : implementation by backprojection
    sum : implementation by summation in real space
    scipy.interpolate.griddata : the used interpolation method
    """
    if jmm is not None:
        jmm.value = 4
        
    if len(sinogram[0]) %2 == 0:
        warnings.warn("Fourier interpolation with slices that have"+
                      " even dimensions leads to image distortions!")
    # projections
    p_x = sinogram

    # Fourier transform of the projections
    # The sinogram data is shifted in Fourier space
    P_fx = np.fft.fft(np.fft.ifftshift(p_x, axes=-1), axis=-1)


    if jmc is not None:
        jmc.value += 1

    # This paragraph could be useful for future version if the reconstruction
    # grid is to be changed. They directly affect *P_fx*.
    # if False:
    #    # Resize everyting
    #    factor = 10
    #    P_fx = np.zeros((len(sinogram), len(sinogram[0])*factor), dtype=np.complex128)
    #    newp = np.zeros((len(sinogram), len(sinogram[0])*factor), dtype=np.complex128)
    #    for i in range(len(sinogram)):
    #        x = np.linspace(0, len(sinogram[0]), len(sinogram[0]))
    #        dint = intp.interp1d(x, sinogram[i])
    #        xn = np.linspace(0, len(sinogram[0]), len(sinogram[0])*factor)
    #        datan = dint(xn)
    #        datfft = np.fft.fft(datan)
    #        newp[i] = datan
    #        # shift datfft
    #        P_fx[i] = np.roll(1*datfft,0)
    #    p_x = newp
    #if False:
    #    # Resize the input image
    #    P_fx = np.zeros(p_x.shape, dtype=np.complex128)
    #    for i in range(len(sinogram)):
    #        factor = 2
    #        x = np.linspace(0, len(sinogram[0]), len(sinogram[0]))
    #        dint = intp.interp1d(x, sinogram[i])#, kind="nearest")
    #        xn = np.linspace(0, len(sinogram[0]), len(sinogram[0])*factor)
    #        datan = dint(xn)
    #        datfft = np.fft.fft(datan)
    #        fftint = intp.interp1d(xn, datfft)
    #        start = (len(xn) - len(x))/2
    #        end = (len(xn) + len(x))/2
    #        fidata = fftint(x)
    #        datfft = np.fft.fftshift(datfft)
    #        datfft = datfft[start:end]
    #        datfft = np.fft.ifftshift(datfft)
    #        P_fx[i] = 1*datfft

    # angles need to be normalized to 2pi
    # if angles star with 0, then the image is falsly rotated
    # unfortunately we still have ashift in the data.
    ang = ( angles.reshape(-1, 1) )
    #ang = ang + ang[1][0]

    # compute frequency coordinates fx
    fx = np.fft.fftfreq(len(p_x[0]))

    #fx = np.linspace(-np.max(fx), np.max(fx), len(p_x[0]))
    fx = fx.reshape(1, -1)
    # fy is zero
    fxl =  (fx)*np.cos(ang)
    fyl =  (fx)*np.sin(ang)
    # now fxl, fyl, and P_fx have same shape
    
    # DEBUG: plot coordinates of positions of projections in fourier domain
    #from matplotlib import pylab as plt
    #plt.figure()
    #for i in range(len(fxl)):
    #    plt.plot(fxl[i],fyl[i],"x")
    #plt.axes().set_aspect('equal')
    #plt.show()


    # flatten everything for interpolation
    Xf = fxl.flatten()
    Yf = fyl.flatten()
    Zf = P_fx.flatten()
    # rintp defines the interpolation grid
    rintp = np.fft.fftshift(fx.reshape(-1))

    if jmc is not None:
        jmc.value += 1
        
    ## The code block yields the same result as griddata (below)
    # interpolation coordinates
    #Rf = np.zeros((len(Xf),2))
    #Rf[:,0] = 1*Xf
    #Rf[:,1] = 1*Yf
    ## reconstruction coordinates
    #Nintp = len(rintp)
    #Xn, Yn = np.meshgrid(rintp,rintp)
    #Rn = np.zeros((Nintp**2, 2))
    #Rn[:,0] = Xn.flatten()
    #Rn[:,1] = Yn.flatten()
    #
    #if intp_method.lower() == "bilinear":
    #    Fr = intp.LinearNDInterpolator(Rf,Zf.real)
    #    Fi = intp.LinearNDInterpolator(Rf,Zf.imag)
    #elif intp_method.lower() == "nearest":
    #    Fr = intp.NearestNDInterpolator(Rf,Zf.real)
    #    Fi = intp.NearestNDInterpolator(Rf,Zf.imag)
    #else:
    #    raise NotImplementedError("Unknown interpolation type: {}".format(
    #                               intp_method.lower()))
    #    return
    #Fcomp = (Fr(Rn) + 1j*Fi(Rn)).reshape(Nintp,Nintp)

    # The actual interpolation
    Fcomp = intp.griddata((Xf, Yf), Zf, (rintp[None,:], rintp[:,None]),
                          method=intp_method)
                          
    if jmc is not None:
        jmc.value += 1
        
    # removed nans
    Fcomp[np.where(np.isnan(Fcomp))] = 0

    f = np.fft.fftshift( np.fft.ifft2(np.fft.ifftshift(Fcomp)) )

    if jmc is not None:
        jmc.value += 1

    return f
    
    
def sum(sinogram, angles, jmc=None, jmm=None):
    u""" 2D sum-reconstruction with the Fourier slice theorem

    Computes the inverse of the radon transform by computing the
    integral in real space.


    Parameters
    ----------
    sinogram : (A,N) ndarray
        Two-dimensional sinogram of line recordings.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally 
        distributed from zero to PI.
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
        

    Returns
    -------
    out : ndarray
        The reconstructed image.


    See Also
    --------
    backproject : implementation by backprojection
    fourier_map : implementation by summation in real space
    """
    if jmm is not None:
        jmm.value = sinogram.shape[1]**2 + 1
    # In the script we used the unitary angular frequency (uaf) Fourier
    # Transform. The discrete Fourier transform is equivalent to the
    # unitary ordinary frequency (uof) Fourier transform.
    #
    # uof: f₁(ξ) = int f(x) exp(-2πi xξ)
    #
    # uaf: f₃(ω) = (2π)^(-n/2) int f(x) exp(-i ωx)
    #
    # f₁(ξ) = (2π)^(n/2) f₃(ω)
    # ω = 2πξ
    #
    # We have a one-dimensional (n=1) Fourier transform and UB in the
    # script is equivalent to f₃(ω). Because we are working with the
    # uaf, we divide by sqrt(2π) after computing the fft with the uof.
    
    # Corresponding sample frequencies
    fx = np.fft.fftfreq(sinogram[0].shape[0]) # 1D array
    kx = 2*np.pi*fx
    
    # Get the angles ϕ₀.
    phi0 = (angles).reshape(-1,1)
    # Differentials for integral
    dphi0 = len(angles)/np.pi
    dkx = np.abs(kx[1] - kx[0])

    #kx = np.fft.fftshift(kx)
    # We will later multiply with phi0.
    # Make sure we are using correct shapes
    kx = kx.reshape(1, -1)


    # Create the integrand
    # Integrals over ϕ₀ [0,2π]; kx [-k,k]
    #   - We do not have double coverage, as the input data is only
    #     from 0 to 180 degrees.
    #   - unitary angular frequency to unitary ordinary frequency
    #     conversion performed in calculation of P=FT(sinogram).
    #
    # f(r) = 1 / ( (2π)^(3/2) )                     (prefactor)
    #      * iint dϕ₀ dkx                           (prefactor)
    #      * |kx|                                   (prefactor)
    #      * P_ϕ₀(kx)                               (dependent on ϕ₀)
    #      * exp( i kx (cos(ϕ₀)*x + sin(ϕ₀)*y ) )  (dependent on ϕ₀ and r)
    #
    # everything that is not dependent on phi0:
    prefactor  = 1 / ( (2*np.pi)**(3/2) )
    prefactor *= dphi0*dkx
    prefactor *= np.abs(kx)
    
    # Initiate function f
    N = len(sinogram[0])
    coords = np.linspace(-N/2.,N/2., N, endpoint=False)
    x, y = np.meshgrid(coords, coords)
    points = np.zeros((N**2, 2))
    points[:,0] = x.flatten()
    points[:,1] = y.flatten()
    f = np.zeros(len(points), dtype=np.complex128)
    lenf = len(f)
    
    # Initiate vector r that corresponds to calculating a value of f.
    r = np.zeros((2,1,1), dtype=np.complex256)
    
    # Create counters for real and imaginary parts of f.
    freal = 0
    fimag = 0

    # Compute the Fourier transform of uB.
    # This is true: np.fft.fft(UB)[0] == np.fft.fft(UB[0])
    # because axis -1 is always used.
    P = np.fft.fft(np.fft.ifftshift(sinogram, axes=-1))/np.sqrt(2*np.pi)

    if jmc is not None:
        jmc.value += 1
    

    for j in xrange(lenf):
        # Get r (We compute f(r) in this for-loop)
        r[0][:] = points[j][0] #x
        r[1][:] = points[j][1] #y

        # Integrand changes with r, so we have to create a new
        # array:
        integrand = prefactor * P
        
        ## Reminder:
        # f(r) = 1 / ( (2π)^(3/2) )                  (prefactor)
        #      * iint dϕ₀ dkx                           (prefactor)
        #      * |kx|                                   (prefactor)
        #      * P_ϕ₀(kx)                               (dependent on ϕ₀)
        #      * exp( i kx (cos(ϕ₀)*x + sin(ϕ₀)*y ) )  (dependent on ϕ₀ and r)
        #
        # everything that is not dependent on phi0:
        integrand *= np.exp(1j*kx*(
                                      r[0] * np.cos(phi0) +
                                      r[1] * np.sin(phi0)   )
                                   )
        
        # Calculate the integral for the position r
        integrand.sort()
        f[j] = np.sum(integrand)
        
        # Display how far we are
        if jmc is not None:
            jmc.value += 1

    return f.reshape((N,N))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
u"""

There are several methods to compute the inverse *Radon* transform.
The module *radontea* implements some of them. I focussed on code
readability and thorough comments. The result is a collection of
algorithms that are suitable for *tea*\ching the basics of
computerized tomography.

"""

from ._Back import *
from ._Back_iterative import *
from ._Back_Fan import *
from ._Radon import *

from ._Back_3D import *


__version__ = "0.1.4"
__author__ = "Paul Mueller"
__email__ = "paul.mueller@biotec.tu-dresden.de"
__license__ = "BSD (3-Clause)"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Give the user a general idea of what this module does.
"""
import numpy as np

from . import _Back as back
from . import _Back_iterative as back_it
from . import _Radon as radon
from ._logo import get_original

# If available, use jobmanager to display progress
try:
    import jobmanager as jm
    for module in [back, back_it, radon]:
        jm.decorators.decorate_module_ProgressBar(module, interval=.1, verbose=1)
except:
    pass
        
        
if __name__ == "__main__":
    import matplotlib
    matplotlib.use("wxagg")
    from matplotlib import pylab as plt
    from matplotlib import cm
        
    N=55
    It=7
    A=13

    angles = np.linspace(0,np.pi,A)
    im = get_original(N)
    
    callback = None
    
    sino = radon.radon(im, angles)
    
    fbp = back.backproject(sino, angles)
    
    fintp = back.fourier_map(sino, angles)
    
    alg = back_it.sart(sino, angles, iterations=It)

    im2 = ( im >= (im.max()/5) ) * 255
    sino2 = radon.radon(im2, angles)
    fbp2 = back.backproject(sino2, angles)
    fintp2 = back.fourier_map(sino2, angles)
    alg2 = back_it.sart(sino2, angles, iterations=It)

    plt.figure(figsize=(15,8))
    
    plt.subplot(2,5,1)
    plt.imshow(im, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('uint16 image')

    plt.subplot(2,5,2)
    plt.imshow(sino, cmap=plt.cm.gray)
    plt.axis('off')
    plt.title('sinogram ({} proj.)'.format(A))

    plt.subplot(2,5,3)
    plt.imshow(fbp, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('backprojection')
    
    plt.subplot(2,5,4)
    plt.imshow(fintp.real, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('Fourier interpolation')


    plt.subplot(2,5,5)
    plt.imshow(alg, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('SART ({} iterations)'.format(It))

    plt.subplot(2,5,6)
    plt.imshow(im2, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('binary image')

    plt.subplot(2,5,7)
    plt.imshow(sino2, cmap=plt.cm.gray)
    plt.axis('off')
    plt.title('sinogram ({} proj.)'.format(A))
    
    plt.subplot(2,5,8)
    plt.imshow(fbp2, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('backprojection')
    
    plt.subplot(2,5,9)
    plt.imshow(fintp2.real, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('Fourier interpolation')

    plt.subplot(2,5,10)
    plt.imshow(alg2, cmap=plt.cm.gray, vmin=0, vmax=255)
    plt.axis('off')
    plt.title('SART ({} iterations)'.format(It))
    
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" _Back_Fan.py

The inverse Radon transform with non-iterative techniques for
a fan-beam geometry.
"""
from __future__ import division

import numpy as np
import scipy

from ._Radon import get_fan_coords

__all__= ["sa_interpolate",
          "lino2sino"]


def lino2sino(linogram, lDS, stepsize=1, det_spacing=1, numang=None,
              retang=False, jmc=None, jmm=None):
    """ Convert linogram to sinogram for an equispaced detector.

    Parameters
    ----------
    linogram : real 2d ndarray of shape (D, A*)
        Linogram from synthetic aperture measurements.
    lDS : float
        Distance from point source to detector in au.
    stepsize : float
        Translational increment of object in au (stepsize in D).
    det_spacing : float
        Distance between detector positions in au.
    numang : int
        Number of equispaced angles, defaults to linogram.shape[1]
    retang : bool
        Return the corresponding angles for the sinogram.
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
        

    Returns
    -------
    sinogram : 2d ndarray of shape (D, A)
        The distortion-corrected sinogram.        
        If retang is True, then the equispaced angles are returned as
        well.


    Notes
    -----
    This function can be used to convert a linogram obtained with
    fan-beam tomography to a sinogram, which then can be reconstructed
    with the backprojection or fourier mapping algorithms.
    
    
    See Also
    --------
    `radontea.radon_fan_translation`
        The forward process.
    `radontea.sa_interpolate`
        Backprojection that uses this function.
    """
    if not linogram is linogram.real:
        raise ValueError("`linogram` must be a real array!")
        
    (D, det_size) = linogram.shape
    
    if numang is None:
        A = det_size #(detector size determines # of angles)
    else:
        A = numang
    
    if jmm is not None:
        jmm.value = D+A
    
    # equispaced angles and corresponding lateral detector positions.
    angles, xang = get_fan_coords(det_size, det_spacing, lDS, A)
    
    uorig = linogram
    lino = np.zeros((D, A))

    # go into the coordinate system of the original data
    xk = np.linspace(xang[0], xang[-1], det_size, endpoint=True)

    # begin equispacing
    for i in range(D):
        #uscaled[i] = scipy.interpolate.spline(xk, uorig[i], xang)
        lino[i] = scipy.interpolate.spline(xk, uorig[i], xang)
        
        if jmc is not None:
            jmc.value += 1

    # begin angular stretching
    for i in range(A):
        # parametrization of the time axis (spatial information)
        xk = np.linspace(-D*stepsize/2+.5,D*stepsize/2-.5, D, endpoint=True)
        alpha = angles[i]

        ## Centering:
        # The object moves 1px per view in D.
        # We need to translate this lateral movement to a shift that
        # depends on the current angle.
        # What is the distance b/w the center of the object (centered at
        # lDS/2) to the axis alpha = 0?
        deltaD = np.tan(alpha) * lDS/2
        #deltaD = np.tan(alpha) * D/2
        #xnew = xk + deltaD
        
        ## Sheering:
        # At larger angles, the object seems bigger on the screen.
        xnew = xk/np.cos(alpha)+deltaD
        lino[:,i] = scipy.interpolate.spline(xk, lino[:,i], xnew)
        
        if jmc is not None:
            jmc.value += 1
                
    sino = np.transpose(lino)[:,::-1]
    if retang:
        return sino, angles + np.pi/2
    else:
        return sino



def sa_interpolate(linogram, lDS, method, stepsize=1, det_spacing=1,
                   numang=None, jmm=None, jmc=None, **kwargs):
    """ 2D synthetic aperture reconstruction
    
    Computes the inverse of the fan-beam radon transform using inter-
    polation of the linogram and one of the inverse algorithms for
    tomography with the Fourier slice theorem.
    
    Parameters
    ----------
    linogram : 2d ndarray of shape (D, A)
        Input linogram from the synthetic aprture measurement.
    lDS : float
        Distance in pixels between source and detector.
    method : callable
        Reconstruction method, e.g. `radontea.backproject`.
    numang : int
        Number of angles to be used for the sinogram. A higher number
        increases quality, but interpolation takes longer. By default
        numang = linogram.shape[1].
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
    **kwargs : dict
        Keyword arguments for `method`.
    
    
    See Also
    --------
    `radontea.radon_fan_translation`
        The forward process.
    `radontea.lino2sino`
        Linogram to sinogram conversion.
    """
    sino, angles = lino2sino(linogram, lDS, numang=numang, retang=True,
                             stepsize=stepsize, det_spacing=det_spacing,
                             jmm=jmm, jmc=jmc)
    #kwargs["jmm"] = jmm
    #kwargs["jmc"] = jmc
    return method(sino, angles, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    _Radon.py
    
    Performs the radon transform.
"""
from __future__ import division
from __future__ import print_function

import numpy as np
import os
import scipy
import scipy.ndimage

import warnings

__all__ = ["radon", "radon_fan_translation", "get_fan_coords",
           "get_det_coords"]



def get_det_coords(size, spacing):
    """ Compute pixel-center positions for 2d detector
    
    The centers of the pixels of a detector are usually not aligned to
    a pixel grid. If we center the detector at the origin, odd images
    will have a pixel at zero, whereas even images will have two pixels
    next to the actual center.
    
    Parameters
    ----------
    size : int
        Size of the detector (number of detection points).
    spacing : float
        Distance between two detection points in pixels.
        
    Returns
    -------
    arr : 1D ndarray
        Pixel coordinates.
    """
    latmax = (size/2-.5) * spacing 
    lat = np.linspace(-latmax, latmax, size, endpoint=True)
    return lat
    
    

def get_fan_coords(size, spacing, distance, numang):
    """ Compute equispaced angular coordinates for 2d detector
    
    The centers of the pixels of a detector are usually not aligned to
    a pixel grid. If we center the detector at the origin, odd images
    will have a pixel at zero, whereas even images will have two pixels
    next to the actual center.
    We want to compute an equispaced array of `numang` angles that go
    between the centers of the outmost far pixels of the detector.
    
    Parameters
    ----------
    size : int
        Size of the detector (number of detection points).
    spacing : float
        Distance between two detection points in pixels.
    distance : float
        Axial distance from the detector to the angular measurement
        position in pixels.
    numang : int
        Number of angles.


    Returns
    -------
    angles, latpos : two 1D ndarrays
        Angles and pixel coordinates at the detector.
        
        
    Notes
    -----
    Actually one would not need to define spacing and distance, but for
    convenience, these parameters are separated and an arbitrary uint
    'pixel' is defined.
    """
    # Compute the angular positions of the outmost pixels
    latmax = (size/2-.5) * spacing
    angmax = abs(np.arctan2(latmax, distance))
    # equispaced angles
    angles = np.linspace(-angmax, angmax, numang, endpoint=True)
    
    latang = np.tan(angles)*distance
    latang[0] = -latmax
    latang[-1] = latmax
    
    return angles, latang



def radon(arr, angles, jmc=None, jmm=None):
    """ Compute the Radon transform (sinogram) of a circular image.


    The `scipy` Radon transform performs this operation on the 
    entire image, whereas this implementation requires an input
    image that has gray-scale values of 0 outside of a circle
    with diameter equal to the image size.


    Parameters
    ----------
    arr : ndarray, shape (N,N)
        the input image.
    angles : ndarray, length A
        angles or projections in radians
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
        

    Returns
    -------
    outarr : ndarray of floats, shape (A, N)
        Sinogram of the input image. The i'th row contains the
        projection data of the i'th angle.


    See Also
    --------
    scipy.ndimage.interpolation.rotate :
        The interpolator used to rotate the image.
    """
    # This function also works with single angles
    angles = np.atleast_1d(angles)
    A = angles.shape[0]
    N = len(arr)
    # The radon function from skimage.transform doeas not allow
    # to reshape the image (one could cut it of course).
    # Furthermore, no interpolation is used or at least I do not
    # know what kind of interpolation is used (_wharp_fast?).
    # outarray: x-axis: projection
    #           y-axis: 
    outarr = np.zeros((A, N))
    # jobmanager
    if jmm is not None:
        jmm.value = A + 1
    if jmc is not None:
        jmc.value += 1
        
    for i in np.arange(A):
        rotated = scipy.ndimage.rotate(arr, angles[i]/np.pi*180, order=3,
                  reshape=False, mode="constant", cval=0) #black corner
        # sum along some axis.
        outarr[i] = rotated.sum(axis=0)
        if jmc is not None:
            jmc.value += 1
    return outarr



def radon_fan_translation(arr, det_size, det_spacing=1, shift_size=1,
                          lS=1, lD=None, return_ang=False,
                          jmc=None, jmm=None):
    """ Compute the Radon transform for a fan beam geometry
        
    In contrast to `radon`, this function uses (1) a fan-beam geometry
    (the integral is taken along rays that meet at one point), and
    (2) translates the object laterally instead of rotating it. The
    result is sometimes referred to as 'linogram'.

    x
    ^ 
    |
    ----> z


    source       object   detector

              /             . (+det_size/2, lD)
    (0,-lS) ./   (0,0)      .
             \              .
              \             . (-det_size/2, lD)


    The algorithm computes all angular projections for discrete
    movements of the object. The position of the object is changed such
    that its lower boundary starts at (det_size/2, 0) and its upper
    boundary ends at (-det_size/2, 0) at increments of `shift_size`.


    Parameters
    ----------
    arr : ndarray, shape (N,N)
        the input image.
    det_size : int
        The total detector size in pixels. The detector centered to the
        source. The axial position of the detector is the center of the
        pixels on the far right of the object.
    det_spacing : float
        Distance between detection points in pixels.
    shift_size : int
        The amount of pixels that the object is shifted between
        projections.
    lS : multiples of 0.5
        Source position relative to the center of `arr`. lS >= 1.
    lD : int
        Detector position relative to the center `arr`. Default is N/2.
    return_ang : bool
        Also return the angles corresponding to the detector pixels.
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
        

    Returns
    -------
    outarr : ndarray of floats, shape (N+det_size,det_size)
        Linogram of the input image. Where N+det_size determines the
        lateral position of the sample.


    See Also
    --------
    scipy.ndimage.interpolation.rotate :
        The interpolator used to rotate the image.
    radontea.radon
        The real radon transform.
    """
    N = arr.shape[0]
    if lD is None:
        lD = N/2
    lS = abs(lS)
    lDS = lS + lD

    numsteps = int(np.ceil((N+det_size)/shift_size))
    
    if jmm is not None:
        jmm.value = numsteps + 2
    
    # First, create a zero-padded version of the input image such that
    # its center is the source - this is necessary because we can
    # only rotate through the center of the image.
    #arrc = np.pad(arr, ((0,0),(N+2*lS-1,0)), mode="constant")    
    if lS <= N/2:
        pad = 2*lS
        arrc = np.pad(arr, ((0,0),(pad,0)), mode="constant")
        # The center is where either b/w 2 pixels or at one pixel,
        # as it is for the input image.
    else:
        pad = N/2 + lS
        if pad%1 == 0: #even
            pad = int(pad)
            # We padd to even number of pixels.
            # The center of the image is between two pixels (horizontally).
            arrc = np.pad(arr, ((0,0),(pad,0)), mode="constant")
        else: #odd
            # We padd to odd number of pixels.
            # The center of the image is a pixel.
            pad = int(np.floor(pad))
            arrc = np.pad(arr, ((0,0),(pad,0)), mode="constant")
    
    # Lateral and axial coordinates of detector pixels
    axi = np.ones(det_size)*lDS
    # minus .5 because we use the center of the pixel
    
    if det_size%2 == 0: #even
        even = True
    else: #odd
        even = False

    if jmc is not None:
        jmc.value += 1

    lat = get_det_coords(det_size, det_spacing)

    angles = np.arctan2(lat,axi)
    
    # Before we can rotate the image for every lateral position of the
    # object, we need to zero-pad the image with the lateral detector
    # size. We pad only the bottom of the image, putting its center at 
    # the upper detector boundary.
    pad2 = det_size
    padset = np.pad(arrc, ((0,pad2), (0,0)), mode="constant")
    A = angles.shape[0]
    lino = np.zeros((numsteps, A))

    if jmc is not None:
        jmc.value += 1
    
    for i in range(numsteps):
        padset = np.roll(padset, shift_size, axis=0)
        # cut out a det_size slice
        curobj = padset[N:]
        for j in range(A):
            ang = angles[j]
            rotated = scipy.ndimage.rotate(curobj, ang/np.pi*180,
                        order=3, reshape=False, mode="constant", cval=0)
            if even:
                warnings.warn("For even images the linogram is of "+
                              "reduced resolution, because the center "+
                              "of the image does not coincide with "+
                              "a center of a pixel.")
                centerid = int(len(rotated)/2)
                lino[i,j] = np.sum(rotated[centerid-1:centerid+1, :])/2
            else: #odd
                centerid = int(np.floor(len(rotated)/2))
                lino[i,j] = np.sum(rotated[centerid, :])
        
        if jmc is not None:
            jmc.value += 1
            
    if return_ang:
        return lino, angles
    else:
        return lino





#!/usr/bin/env python
# To create a distribution package for pip or easy-install:
# python setup.py sdist
from setuptools import setup, find_packages
from os.path import join, dirname, realpath
from warnings import warn

import radontea

name='radontea'

setup(
    name=name,
    author=radontea.__author__,
    author_email=radontea.__email__,
    version=radontea.__version__,
    license=radontea.__license__,
    url='https://github.com/paulmueller/radontea',
    packages=[name],
    package_dir={name: name},
    description='Collection of algorithms to compute the inverse Radon transform',
    long_description=open(join(dirname(__file__), 'README.txt')).read(),
    install_requires=[ "NumPy >= 1.5.1", "SciPy >= 0.8.0"],
    keywords=["tomography", "ct", "radon"],
    extras_require={
                    'doc': ['sphinx']
                   },
    classifiers= [
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Topic :: Scientific/Engineering :: Visualization',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL']
    )




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

_Back_iterative.py

Investigation of non-diffraction tomography methods.

- perform algebraic reconstruction of a sinogram taken from angles
  that are not equidistant

- Algebraic reconstruction technique (ART)

- Simultaneous algebraic reconstruction technique (SART)
  - no hamming window
  - not parallelized

"""
from __future__ import division, print_function

import numpy as np
import os


__all__ = ["art", "sart"]


def art(sinogram, angles, initial=None, iterations=1,
        jmc=None, jmm=None):
    u""" Algebraic Reconstruction Technique
    
    The Algebraic Reconstruction Technique (ART) iteratively
    computes the inverse of the Radon transform in two dimensions.
    The reconstruction technique uses *rays* of the diameter of
    one pixel to iteratively solve the system of linear equations
    that describe the projection process. The binary weighting
    factors are
    
     - 1, if the center of the a pixel is within the *ray*
     - 0, else


    Parameters
    ----------
    sinogram : ndarrayy, shape (A,N)
        Two-dimensional sinogram of line recordings.
    angles : ndarray, length A
        Angular positions of the `sinogram` in radians. The angles
        at which the sinogram slices were recorded do not have to be
        distributed equidistantly as in :func:`backproject`.
        The angles are internaly converted to modulo PI.
    initial : ndarray, shape (N,N), optional
        The initial guess for the solution.
    iterations : int
        Number of iterations to perform.
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
        

    See Also
    --------
    sart : simultaneous algebraic reconstruction technique


    Notes
    -----
    For theoretical backround, see
    Kak, A. C., & Slaney, M.. *Principles of Computerized
    Tomographic Imaging*, SIAM, (2001)
    
    Sec. 7.2:
    *"ART reconstrutions usually suffer from salt and pepper noise,
    which is caused by the inconsitencies introuced in the set of
    equations by the approximations commonly used for* 
    :math:`w_{ik}` *'s."*
    """
    # make sure `iterations` is an integer
    iterations = int(iterations)
    N = sinogram.shape[1]
    A = angles.shape[0]
    # jobmanager
    if jmm is not None:
        jmm.value = A*iterations + 1
    # Meshgrid for weigths
    center = N/2.0
    x = np.arange(N) - center +.5
    X, Y = np.meshgrid(x,x)
    # initiate array
    if initial is None:
        f = np.zeros((N,N),dtype=np.dtype(float))
    else:
        f = 1*initial.transpose()[::-1]
    # Make sure all angles are in [0,PI)
    for i in np.arange(A):
        if angles[i] > np.pi:
            offset = np.floor(angles[i]/np.pi)
            angles[i] -= offset*np.pi
        elif angles[i] < 0:
            offset = np.floor(np.abs((angles[i]/np.pi)))+1
            angles[i] += offset*np.pi
        if angles[i] == np.pi:
            angles[i] = 0

    # These lambda functions return the two x- and y- values of two
    # points projected onto a line along the x- and y-axis, having the
    # angle angle_k in radians.
    # Sourcing this out to here makes things a little faster.
    def GetLambdaLines(angle_k):
        if angle_k == 0:
            # Divide by zero error for tan
            # We have horizontal lines (parallel to x)
            line1 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: (x2-.5, ypi1*np.ones(y2.shape))
            line2 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: (x2+.5, ypi2*np.ones(y2.shape))
        elif angle_k == np.pi/2:
            # We have vertical lines (parallel to y)
            line1 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: (xpi1*np.ones(x2.shape), y2+.5)
            line2 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: (xpi2*np.ones(x2.shape), y2-.5)
        elif angle_k < np.pi/2:
            # CASE 1
            # Compute any other positions on the lines from the given things
            line1 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: ( (y2-ypi1)/np.tan(angle_k) + xpi1,
                                    (x2-xpi1)*np.tan(angle_k) + ypi1)
                #def line2(x2,y2):
                #    y = (x2-xpi2)/np.tan(angle_k) - ypi2
                #    x = (y2-ypi2)*np.tan(angle_k) - xpi2
                #    return x,y
            line2 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: ( (y2-ypi2)/np.tan(angle_k) + xpi2,
                                    (x2-xpi2)*np.tan(angle_k) + ypi2)
        else:
            # CASE 2: Switch x-output - only for speed. Not very obvious - possibly a hack.
            # Compute any other positions on the lines from the given things
            line1 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: ( (y2-ypi2)/np.tan(angle_k) + xpi2,
                                    (x2-xpi1)*np.tan(angle_k) + ypi1)
                #def line2(x2,y2):
                #    y = (x2-xpi2)/np.tan(angle_k) - ypi2
                #    x = (y2-ypi2)*np.tan(angle_k) - xpi2
                #    return x,y
            line2 = lambda x2,y2,xpi1,ypi1,xpi2,ypi2,angle_k: ( (y2-ypi1)/np.tan(angle_k) + xpi1,
                                    (x2-xpi2)*np.tan(angle_k) + ypi2)
        return line1, line2
    # Sort angles?
    # This could increase the speed of convergence.    

    # f[j] is consistent with Kak, Slaney
    f = f.flatten()

    if jmc is not None:
        jmc.value += 1

    for iteration in np.arange(iterations):
        #
        # i iterates the rays
        # j iterates the flattened outarr
        # k iterates the angles
        #
        for k in np.arange(A):
            # From now on we work in radians
            angle_k = angles[k]
            # p[i] is consistent with Kak, Slaney
            p = sinogram[k]
               # # We do not store the binary weights for each ray:
               # # For large images this thing could get big:
               # w = np.zeros((len(p),N*N), dtype=np.bool)
               # # w[i][j]
            line1, line2 = GetLambdaLines(angle_k)
            # CASES for all 2 quadrants. Angles are modulo PI
            # This case stuff is dependent on the angle. We enumerate
            # pr in negative mathematical angular direction
            #
            # This is the position on the projection, centered around 0:
            pr = np.arange(len(p)) - center + .5 # radial distance,
            #
            if angle_k <= np.pi/2:
                # 0 to PI/2
                case = 1
                # position of each p[i] in the centered *outarr*.
                x_p1 = (pr-.5)*np.sin(angle_k)                # a position of line1 in X
                y_p1 = -(pr-.5)*np.cos(angle_k)                # a position of line1 in Y
                x_p2 = (pr+.5)*np.sin(angle_k)                # a position of line2 in X
                y_p2 = -(pr+.5)*np.cos(angle_k)                # a position of line2 in Y
            else:
                # PI/2 to PI
                # position of each p[i] in the centered *outarr*.
                x_p1 = (pr+.5)*np.sin(angle_k)                # a position of line1 in X
                y_p1 = -(pr+.5)*np.cos(angle_k)                # a position of line1 in Y
                x_p2 = (pr-.5)*np.sin(angle_k)                # a position of line2 in X
                y_p2 = -(pr-.5)*np.cos(angle_k)                # a position of line2 in Y
                case = 2
            for i in np.arange(len(p)):
                # If the angle is zero, then we are looking at the
                # projections onto the right side. The indices are
                # enumerated from bottom to top.
                w_i = np.zeros((N,N), dtype=np.bool)
                       # # where does the ray cut w_i?
                       # xpi1 = x_p1[i]
                       # ypi1 = y_p1[i]
                       # xpi2 = x_p2[i]
                       # ypi2 = y_p2[i]
                # Check if a value is between the two lines (within ray)
                # For each value of X and Y, compute what the values of the
                # line would be.
                xl1, yl1 = line1(X,Y,x_p1[i],y_p1[i],x_p2[i],y_p2[i],angle_k)
                xl2, yl2 = line2(X,Y,x_p1[i],y_p1[i],x_p2[i],y_p2[i],angle_k)
                #
                AND = np.logical_and
                # Values lie between the two lines if the following is True:
                # Furthermore we restrict ourselves to a disk.
                w_i = AND(X**2 + Y**2 < center**2,
                          AND(AND(xl1 < X ,X < xl2),
                              AND(yl1 > Y ,Y > yl2))).flatten()
                #
                # i iterates the rays
                # j iterates the flattened outarr
                # k iterates the angles
                ## In each iteration of the angles the image is changed
                #if np.sum(w_i) != 0:
                #    f += ( p[i] - np.sum(f*w_i) )/np.sum(w_i) * w_i
                # This is faster because we don't a the zeros.
                f[np.where(w_i==True)] += np.divide(
                        ( p[i] - np.sum(f[np.where(w_i==True)]) ),
                        np.sum(w_i))
                del w_i
                #
                 # # Export single steps as bmp files...
                 # #if i%10 == 1:
                 # #    image = f.reshape((N,N))[::-1]
            if jmc is not None:
                jmc.value += 1
                 # #    proc_arr2im(image, cut=True).save(os.path.join(DIR,"test/Elephantulus_small_art_%02d_%04d_%08d.bmp" % (iteration,k,i)))
    # By slicing in-place [::-1] we get rid of the inversion of the
    # image along the y-axis.
    
    return f.reshape((N,N))[::-1].transpose()


def sart(sinogram, angles, initial=None, iterations=1,
         jmc=None, jmm=None):
    u""" Simultaneous Algebraic Reconstruction Technique
    
    
    SART computes an inverse of the Radon transform in two dimensions.
    The reconstruction technique uses "rays" of the diameter of
    one pixel to iteratively solve the system of linear equations
    that describe the image. The weighting factors are bilinear
    elements. At the beginning and end of each ray, only partial
    weights are used. The pixel values of the image are updated
    only after each iteration is complete.
    
    
    Parameters
    ----------
    sinogram : ndarrayy, shape (A,N)
        Two-dimensional sinogram of line recordings.
    angles : ndarray, length A
        Angular positions of the `sinogram` in radians. The angles
        at which the sinogram slices were recorded do not have to be
        distributed equidistantly as in backprojection techniques.
        The angles are internaly converted to modulo PI.
    initial : ndarray, shape (N,N), optional
        the initial guess for the solution.
    iterations : integer
        Number of iterations to perform.
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.

        
    
    See Also
    --------
    art : algebraic reconstruction technique


    Notes
    -----
    Algebraic reconstruction technique (ART) (see `art`):
        Iterations are performed over each ray of each projection.
        Weighting factors are binary (1 if center of pixel is
        within ray, 0 else). This leads to salt and pepper noise.
    
    Simultaneous iterative reconstruction technique (SIRT):
        Same idea as ART, but for each iteration, the change of the
        image f is computed for all rays and projections separately
        and the weights are applied simultaneously after each
        iteration. The result is a slower convergence but the final
        image is also less noisy.
    
    This implementation does NOT use a hamming window to filter
    the data and to emphasize points at the center of the recon-
    struction region.
    
    For theoretical backround, see
    Kak, A. C., & Slaney, M.. *Principles of Computerized
    Tomographic Imaging*, SIAM, (2001)
    
    Sec 7.4:
    *"[SART] seems to combine the best of ART and SIRT. [...] Here
    are the main features if SART: First, [...] the traditional
    pixel basis is abandonded in favor of bilinear elements
    [e.g. interpolation]. Also, for a circular reconstruction
    region, only partial weights are assigned to the first and last
    picture elements on the individual rays. To further reduce the
    noise [...], the correction terms are simultaneously applied
    for all the rays in one projection [...]."*
    """
    # make sure `iterations` is an integer
    iterations = int(iterations)
    N = sinogram.shape[1]
    A = angles.shape[0]
    # jobmanager
    if jmm is not None:
        jmm.value = A*iterations + 1
    # Meshgrid for weigths
    center = N/2.0
    x = np.arange(N) - center +.5
    # initiate array
    if initial is None:
        g = np.ones((N,N),dtype=np.dtype(float))
    else:
        g = 1*initial[::-1]
    # Make sure all angles are in [0,PI)
    for i in np.arange(A):
        if angles[i] > np.pi:
            offset = np.floor(angles[i]/np.pi)
            angles[i] -= offset*np.pi
        elif angles[i] < 0:
            offset = np.floor(np.abs((angles[i]/np.pi)))+1
            angles[i] += offset*np.pi
        if angles[i] == np.pi:
            angles[i] = 0
            
    # g^k --> g^[iteration] is consistent with Kak, Slaney
    # k is the number of the current iteration
    # i iterates the rays of the projected image (1 to len(p))
    # j iterates over the entire flattened g
    # A[i,j] holds the weights for each ray i for all? angles l.
    # a[i] denotes the ith row vector of the array A[i,j]
    
    # g^[iteration] += a[i]*(p[i]-a[i]*g^[iteration])/(a[i]*a[i])
    
    g = g.flatten()
    
    if jmc is not None:
        jmc.value += 1

    for k in np.arange(iterations):
        #
        # k iterates the iterations
        # i iterates the rays
        # j iterates the flattened outarr
        # l iterates the angles
        # m iterates the points from which each projection p[i]
        #   is computed
        #
        # initiate array
        dgall = np.zeros((A, N, N))
        for l in np.arange(A):
            # Differences accumulated over one angle
            #dg = np.zeros((N,N), dtype=np.float64)
            dg = dgall[l]
            ai_sum = np.zeros((N,N))
            
            # From now on we work in radians
            angle_l = angles[l]
            # p[i] is consistent with Kak, Slaney
            p = sinogram[l]
            ## Calculate the weights for each p[i] at the current angle
            # p[i] = sum(g[j]*A[i,j])
            # A[i,j] = sum(d[i,j,m]*ds)
            # ds - distance between points, we let it default to half
            # a pixel size.
            
            # We will go through the procedure of calculating the
            # d[i,j,m]'s and A[i,j]'s.
            #print "angle %d" % l
            
            for i in np.arange(N):
                # Get the coordinates m at which we want to get the 
                # interpolations.
                # The resolution is equal to the input resolution
                # Crop the line according to its length in the
                # circular shaped area:
                # Image radius
                #ir = (N+1)/2.
                ir = center
                # x is centered
                dist = np.sqrt(ir**2-x[i]**2)
                # Delta s: distance between sample points
                # Setting it to half the pixel spacing:
                length = 2.*dist
                num_points = 2*np.ceil(2*dist)+1
                # endpoints are on the reconstruction circle.
                # Therefore "-1"
                ds = length/(num_points-1)
                
                # x-values of the line in the non-rotated coordinate
                # system:
                xline = np.linspace(-dist,dist,num_points,endpoint=True)
                # y-values are constant in the non-rotated system
                yline = x[i] * np.ones(len(xline))
                
                # Rotate the coordinate system of the line
                # These are the actual coordinates in the reconstructed
                # image that where used to sum up p[i]
                xrot = xline*np.cos(angle_l) - yline*np.sin(angle_l)
                yrot = xline*np.sin(angle_l) + yline*np.cos(angle_l)

                # Bispline interpolation at the points xrot and yrot
                # This for-loop slows down computation time. We replaced
                # it (see below). The replacement is a little cryptic.
                # Therefore we keep it here.
                    #dijm = np.zeros((N,N,num_points))
                    #for m in np.arange(num_points):
                    #    # get the four pixels that surround the
                    #    # index positions g[px][py]
                    #    px = xrot[m] + center -.5
                    #    py = yrot[m] + center -.5
                    #    
                    #    # According to quadrants
                    #    P1 = [min(np.ceil(px),N-1), min(np.ceil(py),N-1)]
                    #    P2 = [np.floor(px), min(np.ceil(py),N-1)]
                    #    P3 = [np.floor(px), np.floor(py)]
                    #    P4 = [min(np.ceil(px),N-1), np.floor(py)]
                    #    
                    #    # print px, P1[0], P2[0], P3[0], P4[0]
                    #    # Calculate the weigths ai
                    #    
                    #    # See wikipedia article
                    #    # f(0,0) is at P3
                    #    prelx = px-P3[0]
                    #    prely = py-P3[1]
                    #    ## f(1,1)
                    #    dijm[P1[0]][P1[1]][m] = prelx*prely
                    #    ## f(0,1)
                    #    dijm[P2[0]][P2[1]][m] = (1-prelx)*prely
                    #    ## f(0,0)
                    #    dijm[P3[0]][P3[1]][m] = (1-prelx)*(1-prely)
                    #    ## f(1,0)
                    #    dijm[P4[0]][P4[1]][m] = prelx*(1-prely)
                    # ai = ds*np.sum(dijm,axis=2)
                    
                # We spare us the summation over m by summing directly
                dij = np.zeros((N,N),dtype=np.float64)
                
                px = xrot + center -.5
                py = yrot + center -.5
                
                P1x = np.uint64(np.ceil(px)*(np.ceil(px) < N-1) + (N-1)*(np.ceil(px) >= N-1))
                P1y = np.uint64(np.ceil(py)*(np.ceil(py) < N-1) + (N-1)*(np.ceil(py) >= N-1))
                P2x = np.uint64(np.floor(px)*(np.floor(px) > 0))
                P2y = np.uint64(np.ceil(py)*(np.ceil(py) < N-1) + (N-1)*(np.ceil(py) >= N-1))
                P3x = np.uint64(np.floor(px)*(np.floor(px) > 0))
                P3y = np.uint64(np.floor(py)*(np.floor(py) > 0))
                P4x = np.uint64(np.ceil(px)*(np.ceil(px) < N-1) + (N-1)*(np.ceil(px) >= N-1))
                P4y = np.uint64(np.floor(py)*(np.floor(py) > 0))

                prelx = px-P3x
                prely = py-P3y

                Px = np.array([P1x, P2x, P3x, P4x]).flatten()
                Py = np.array([P2y, P2y, P3y, P4y]).flatten()
                
                Pdelta = np.array([ prelx*prely,
                                    (1-prelx)*prely,
                                    (1-prelx)*(1-prely),
                                    prelx*(1-prely)]).flatten()
                
                # Add a factor of 0.5 for points at beginning and end
                Pdelta[[0, len(prelx)-1, len(prelx), 2*len(prelx)-1,
                       2*len(prelx), 3*len(prelx)-1, 3*len(prelx),
                       4*len(prelx)-1]] *= 0.5
                
                # With the help of the StackOverflow.com community
                # we got rid of the for loop over m:
                # convert yourmulti-dim indices to flat indices
                flat_idx = np.ravel_multi_index((Px, Py), dims=dij.shape)
                # extract the unique indices and their position
                unique_idx, idx_idx = np.unique(flat_idx, return_inverse=True)
                # Aggregate the repeated indices 
                deltas = np.bincount(idx_idx, weights=Pdelta)
                # Sum them to your array
                dij.flat[unique_idx] += deltas

                ai = ds*dij
                del dij
                
                ai_dot = np.sum(np.square(ai))

                # So for each pixel that had just contributed, we will
                # add a value to the difference array dg.
                dg += ai * ( p[i] - np.sum(ai.flatten()*g) )/ai_dot

                ai_sum += ai

                del ai
                del ai_dot
            # This illustrates how the black ring forms on the outside.
            # If a low resolution 128x128 is used, one can see the
            # pixelation error.
            #proc_arr2im((g+dg.flatten()).reshape((N,N))[::-1], cut=False).save(os.path.join(DIR,"test/%08d.bmp" % l))
            #proc_arr2im((ai_sum*255/np.max(ai_sum)).reshape((N,N))[::-1], cut=False).save(os.path.join(DIR,"test/ai%08d.bmp" % l))
            dg[np.where(ai_sum != 0)] /= ai_sum[np.where(ai_sum != 0)]
            del ai_sum
            if jmc is not None:
                jmc.value += 1

        ## Only apply the average from all differences dgall
        ## ->leads to slower convergence then ART, but is more accurate
        dgall=np.average(dgall,axis=0).flatten()
        g += dgall
        ##g += dg/ai_sum

        del dgall
         # # Export single steps as bmp files...
         # #if i%10 == 1:
         # #    image = f.reshape((N,N))[::-1]
         # #    proc_arr2im(image, cut=True).save(os.path.join(DIR,"test/Elephantulus_small_art_%02d_%04d_%08d.bmp" % (iteration,k,i)))
         
    # compare input and output:
    #proc_arr2im(g.reshape((N,N)), cut=False).save(os.path.join(DIR,"sart_proc.bmp"))
    #proc_arr2im(initial[::-1], cut=False).save(os.path.join(DIR,"sart_ref.bmp"))

    # By slicing in-place [::-1] we get rid of the inversion of the
    # image along the y-axis.
    
    return g.reshape((N,N))[::-1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    _logo.py
    
    Functions that are used to create the logo of radontea.
"""
from __future__ import division

import numpy as np

if __name__ == "__main__":
    from _Radon import radon
else:
    from ._Radon import radon


def logo(x,y,N):
    """
        Vector representation of radonteach logo.
        
        
        Parameters
        ----------
        x,y : ndarray or float
            coordinates where to calculate the logo.
        N : float
            total size of the image (NxN)
            
            
        Returns
        -------
        v : ndarray
            inverted values of the logo at the coordinates (x,y)
            normalized to one.
    """
    z1 = np.exp(-((x-N/8)**2+(y+N/14)**2)/(N/8)**2)
    z2 = np.exp(-((x+N/10)**2+(y-N/10)**2)/(N/8)**2)

    z3 = np.exp(-((x-N/8)**2+(y+N/10)**2)/(N/12)**2)
    z4 = np.exp(-((x+N/10)**2+(y-N/14)**2)/(N/12)**2)

    z5 = np.exp(-((x)**2+(y)**2)/(N/4)**2)
    z = (z1+z2 -(z3+z4))
    z = (z-z.min())*z5
    z /= z.max()
    z[np.where(x**2+y**2>=(N/2)**2)] = 0
    
    return z


def get_original(N=64):
    """
        Returns a two-dimensional square image that is the image used
        to create the sinogram for the radontea logo.
    """
    x = np.linspace(-N/2, N/2, N, endpoint=False)
    X = x.reshape(1,-1)
    Y = x.reshape(-1,1)
    z = logo(X,Y,N)
    
    return np.array((z)*255, dtype=np.uint16)


def get_logo(N=64):
    """
        Returns a two-dimensional square logo of resolution NxN. This
        function discretizes the vector image representation of the
        radonteach logo.
    """
    a=(get_original(N=N))
    angles = np.linspace(0, np.pi*1.6, N) 
    sinogram = radon(a, angles)
    sinogram = (1-sinogram/sinogram.max())*255
    return sinogram.transpose()


def main():
    # Show the logo
    from matplotlib import pylab as plt
    from matplotlib import cm
    logo = get_logo(N=128)
    plt.imshow(logo, cmap=cm.gray)
    plt.axis("off")
    plt.tight_layout()
    plt.show()
    
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    _Back_3D.py

    Generates 3D variants of the 2D sinogram inversion methods of
    `radontea` using `multiprocessing`.
"""
## TODO
#
# Jobmanager:
# - Local Manager?
# - Example where a function calls a number of other functions with
#   only one Python call and returning a result?
#
from __future__ import division, print_function

import multiprocessing as mp
import numpy as np
import os
import scipy.interpolate as intp
import warnings

__all__= ["backproject_3d", "fourier_map_3d", "sart_3d"]

from ._Back import *
from ._Back_iterative import *


def back_3d(sinogram=None, angles=None, method="backproject", 
            jmc=None, jmm=None, **kwargs):
    u""" 3D inverse with the Fourier slice theorem
    
    Computes the slice-wise 3D inverse of the radon transform using
    multiprocessing.


    Parameters
    ----------
    sinogram : ndarray, shape (A,N,M)
        Three-dimensional sinogram of line recordings. The dimension `M`
        iterates through the slices. The rotation takes place through
        the second (y) axis.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI. Can also be given in `kwargs`.
    method : str
        The method to be used for sinogram inversion (e.g. "sart",
        "backproject", or "foutier_map"). `method` must be a method
        in `radontea`. This parameter is automatically overwritten
        with `radontea.METHOD` unless you are explicitly
        calling `radontea._Back_3D.back_3d`.
    jmc, jmm : instance of `multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        `jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
    *kwargs : dict
        Keyword arguments to `radontea.METHOD`.

    Returns
    -------
    out : ndarray
        The reconstructed image.

    """
    assert sinogram.shape[0] == angles.shape[0], \
            "First dimension of `sinogram` must match size of `angles`"
    assert len(sinogram.shape) == 3, \
            "`sinogram` must have three dimensions."
   
    # Check if the method exists
    if isinstance(method, str):
        try:
            func = globals()[method]
        except:
            raise ValueError("Unknown method: '{}'".format(method))
    
    # Write method to globals so we can wrap it
    _set_method(func.__name__)
    
    
    if angles is not None:
        kwargs["angles"] = angles
    
    (A,N,M) = sinogram.shape
    
    sinogram = sinogram.transpose(0,2,1)
    
    # How long will the algorithm run? - `jobmanager` counters.
    #kwargs["jmc"] = jmc
    #if func.__name__ == "backproject":
    #    if jmm is not None:
    #        jmm.value = M * (A+1) 

    # arguments of the function
    func_args = func.__code__.co_varnames[:func.__code__.co_argcount]
    # default keyword arguments
    func_def = func.__defaults__[::-1]
    
    # build arguments reversely
    arglistsino = list()
    for m in range(M):
        arglist = list()
        kwargs["sinogram"] = sinogram[:,:,m]
        for i, a in enumerate(func_args[::-1]):
            # first set default
            if i < len(func_def):
                val = func_def[i]
            if a in kwargs:
                val = kwargs[a]
            arglist.append(val)
        arglistsino.append(arglist[::-1])
    
    # Mapsinogram to cpus
    # Splice sinogram
    p = mp.Pool(mp.cpu_count())
    
    result = p.map_async(_wrapper_func, arglistsino).get()
    
    shape = (result[0].shape[0], result[0].shape[1], len(result))
    data = np.zeros(shape)
    for m in range(M):
        data[:,:,m] = result[m].transpose()
    data = data.transpose(1,2,0)
    
    return data

    
def _wrapper_func(args):
    # get the method
    func = globals()[_used_method]
    return func(*args)


def _set_method(method):
    global _used_method
    _used_method = method


def _generate_3dfunc(method3d):
    method = method3d[:-3]
    def _generated(*args, **kwargs):
        # translate args to kwargs
        func_args = back_3d.__code__.co_varnames[:back_3d.__code__.co_argcount]
        for i, arg in enumerate(args):
            kwargs[func_args[i]] = arg
        kwargs["method"] = method
        return back_3d(**kwargs)
    _generated.__doc__ = back_3d.__doc__.replace("METHOD", method)
    return _generated
    

_used_method = "none"


glob = globals()
for method3d in __all__:
    glob[method3d] = _generate_3dfunc(method3d)
